from . import scicomap
from . import datasets
from . import cmath
from . import cblind

__version__ = "0.4"
__all__ = ["scicomap", "datasets", "cmath", "cblind", "utils"]

# bound to upper level
from .scicomap import *
from .datasets import *
from .cmath import *
from .cblind import *


# Author declaration
__author__ = "Thomas Bury"



"""
Module collecting function for evaluating and transforming cmap to render
color weak or blind vision
"""

import itertools
import matplotlib.pyplot as plt
import numpy as np
from colorspacious import cspace_converter
from matplotlib.colors import ListedColormap
from scicomap.cmath import get_ctab

_deuter50_space = {"name": "sRGB1+CVD",
                   "cvd_type": "deuteranomaly",
                   "severity": 50}
_deuter50_to_sRGB1 = cspace_converter(_deuter50_space, "sRGB1")

_deuter100_space = {"name": "sRGB1+CVD",
                    "cvd_type": "deuteranomaly",
                    "severity": 100}
_deuter100_to_sRGB1 = cspace_converter(_deuter100_space, "sRGB1")

_prot50_space = {"name": "sRGB1+CVD",
                 "cvd_type": "protanomaly",
                 "severity": 50}
_prot50_to_sRGB1 = cspace_converter(_prot50_space, "sRGB1")

_prot100_space = {"name": "sRGB1+CVD",
                  "cvd_type": "protanomaly",
                  "severity": 100}
_prot100_to_sRGB1 = cspace_converter(_prot100_space, "sRGB1")

_trit50_space = {"name": "sRGB1+CVD",
                 "cvd_type": "tritanomaly",
                 "severity": 50}
_trit50_to_sRGB1 = cspace_converter(_trit50_space, "sRGB1")

_trit100_space = {"name": "sRGB1+CVD",
                  "cvd_type": "tritanomaly",
                  "severity": 100}
_trit100_to_sRGB1 = cspace_converter(_trit100_space, "sRGB1")


def colorblind_transform(RGBA, colorblind_space):
    # clipping, alpha handling
    RGB = RGBA[..., :3]
    RGB = np.clip(colorblind_space(RGB), 0, 1)
    return np.concatenate((RGB, RGBA[..., 3:]), axis=-1)


def _get_color_weak_cmap(color_map, n_images):
    _deuter50_transform = lambda x: colorblind_transform(x, _deuter50_to_sRGB1)
    _deuter100_transform = lambda x: colorblind_transform(x, _deuter100_to_sRGB1)
    _prot50_transform = lambda x: colorblind_transform(x, _prot50_to_sRGB1)
    _trit100_transform = lambda x: colorblind_transform(x, _trit100_to_sRGB1)

    deuter50_cm = _colorblind_cmap(color_map, _deuter50_transform)
    deuter100_cm = _colorblind_cmap(color_map, _deuter100_transform)
    prot50_cm = _colorblind_cmap(color_map, _prot50_transform)
    trit100_cm = _colorblind_cmap(color_map, _trit100_transform)
    c_maps = [color_map, deuter50_cm, prot50_cm, deuter100_cm, trit100_cm]
    n_blank = n_images - 1
    sub_title = ["Normal (~95% of pop)"] + n_blank * [""] + \
                ["Deuter-50% (RG-weak, D/P: 5% of male)"] + n_blank * [""] + \
                ["Prot-50% (RG-weak, D/P: 5% of male)"] + n_blank * [""] + \
                ["Deuter-100% (RG-blind, D/P: g5% of male)"] + n_blank * [""] + \
                ["Trit-100% (BY deficient, very rare)"] + n_blank * [""]

    return c_maps, sub_title


def _get_color_weak_ctab(color_map, n_blank):
    _deuter100_transform = lambda x: colorblind_transform(x, _deuter100_to_sRGB1)
    ctab = get_ctab(color_map)  # get the colormap as a color table in sRGB
    ctab_deuter100 = _deuter100_transform(ctab)

    _deuter50_transform = lambda x: colorblind_transform(x, _deuter50_to_sRGB1)
    ctab_deuter50 = _deuter50_transform(ctab)

    _prot50_transform = lambda x: colorblind_transform(x, _prot50_to_sRGB1)
    ctab_prot50 = _prot50_transform(ctab)

    _trit100_transform = lambda x: colorblind_transform(x, _trit100_to_sRGB1)
    # _trit100_transform = lambda x: colorblind_transform(x, _trit100_to_sRGB1)
    ctab_trit100 = _trit100_transform(ctab)

    c_tabs = [ctab, ctab_deuter50, ctab_prot50, ctab_deuter100, ctab_trit100]
    sub_title = ["Normal (~95% of pop)"] + n_blank * [""] + \
                ["Deuter-50% (RG-weak, D/P: 5% of male)"] + n_blank * [""] + \
                ["Prot-50% (RG-weak, D/P: 5% of male)"] + n_blank * [""] + \
                ["Deuter-100% (RG-blind, D/P: 5% of male)"] + n_blank * [""] + \
                ["Trit-100% (BY deficient, very rare)"] + n_blank * [""]

    return c_tabs, sub_title


def _colorblind_cmap(cmap, c_space_transf):
    ctab = get_ctab(cmap)  # get the colormap as a color table in sRGB
    ctab_cb = c_space_transf(ctab)
    return ListedColormap(np.clip(ctab_cb, 0, 1))


def colorblind_vision(cmap, figsize=None, n_colors=10, facecolor="black"):
    """
    Plot the colormap in different color-weak/blind configuration

    :param cmap: Colormap or ListedColormap object
        the color map you want to draw
    :param figsize: 2-uple of int
        the figure size
    :param n_colors: int, default=10
        the number of colors to plot (e.g. 10 for qualitative and 256 for continuous)
    :return:
    """
    gradient = np.linspace(0, 1, n_colors)
    gradient = np.vstack((gradient, gradient))

    cmap_dic = {}
    visible_spectrum_cmap = plt.get_cmap('gist_rainbow')
    spectral_list, _ = _get_color_weak_cmap(color_map=visible_spectrum_cmap, n_images=0)
    cmap_dic['visible spectrum'] = spectral_list

    if isinstance(cmap, list):
        for cm in cmap:
            cmap_list, _ = _get_color_weak_cmap(color_map=cm, n_images=0)
            cmap_dic[cm.name] = cmap_list
    else:
        cmap_list, _ = _get_color_weak_cmap(color_map=cmap, n_images=0)
        cmap_dic[cmap.name] = cmap_list

    sub_titles = ["Normal\n~95$%$ of pop"] + \
                 ["Deuter-50$%$\nRG-weak, D/P: 5$%$ of male"] + \
                 ["Prot-50$%$\nRG-weak, D/P: 5$%$ of male"] + \
                 ["Deuter-100$%$\nRG-blind, D/P: 5$%$ of male"] + \
                 ["Trit-100$%$\nBY deficient, very rare"]

    nrows = len(spectral_list)
    ncols = len(cmap_dic)

    if figsize is None:
        figsize = (10, 0.25 * nrows)

    fontcolor = "white" if facecolor == "black" else "black"
    fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, facecolor=facecolor)
    fig.subplots_adjust(top=0.95, bottom=0.01, left=0.2, right=0.99)

    cmap_list = list(cmap_dic.values())
    cmap_name = list(cmap_dic.keys())

    for i, j in itertools.product(range(nrows), range(ncols)):
        ax = axes[i, j]
        cmap = cmap_list[j][i]
        ax.imshow(gradient, aspect="auto", cmap=cmap)
        if i == 0:
            font = {'color': fontcolor, 'size': 24}
            ax.set_title(cmap_name[j], fontdict=font)
        if j == 0:
            font = {'color': fontcolor, 'size': 14}
            pos = list(ax.get_position().bounds)
            x_text = pos[0] - 0.01
            y_text = pos[1] + pos[3] / 2.0
            fig.text(x_text, y_text, sub_titles[i], va="center", ha="right", fontdict=font)

    for ax in axes:
        ax[0].set_axis_off()
        ax[1].set_axis_off()
    return fig



"""
Main module for scientific colormaps. It uses the CAM02-UCS color space
Its three coordinates are usually denoted by J', a', and b' and its cylindrical coordinates are J', C', and h'.
This package is built on matplotlib, colorspacious, viscm and EHTplot packages for
the color mathematics and transformations and uses colormaps coming from different packages aiming
to provide scientific colormaps but requiring some adjustments.

 - Provide sequential, multi-sequential, diverging, circular and qualitative (discrete) cmaps
 - Uniformize: linearize the CAM02-UCS lightness J'
 - Symmetrize: make the CAM02-UCS chroma C' symmetrical, bitonic or not, smooth or not
 - Get the matplotlib cmap object, before and after the adjustments
 - Charts to assess the quality of the colormaps (JCh plot)
 - Charts to assess the readability by colour weak/deficient/blind people
 - Charts for illustrating all the available colormaps

The module structure is the following:
---------------------------------------
- ``SciCoMap`` Parent class for all the colormap types
- ``ScicoSequential`` Child class for sequential colormaps
- ``ScicoMultiSequential`` Child class for multi-sequential colormaps
- ``ScicoDiverging`` Child class for diverging colormaps
- ``ScicoCircular`` Child class for circular colormaps (circular diverging and circular flat aka phase)
- ``ScicoMiscellaneous`` Child class for continuous colormaps which are none of the above
- ``ScicoQualitative`` Child class for discrete colormaps
- ``plot_colormap`` function for illustrating all the color maps of a given type
- ``plot_colorblind_vision`` function for comparing the rendering with different color deficiencies
- ``compare_cmap`` function for comparing the rendering when using an image.

"""

import matplotlib.pyplot as plt
import matplotlib.image as mpimg
from matplotlib.colors import ListedColormap
import numpy as np
from scicomap.datasets import load_hill_topography, load_scan_image, load_pic

# Scientific Colours
import colorcet as cc
import cmasher as cmr
from cmcrameri import cm as scico
import cmocean
from palettable.cubehelix import perceptual_rainbow_16, classic_16
from palettable.cartocolors.qualitative import Bold_10, Pastel_10, Prism_10, Vivid_10
from palettable.colorbrewer.qualitative import Set1_9

# internal import
from scicomap.cmath import uniformize_cmap, symmetrize_cmap, unif_sym_cmap, \
    _ax_cylinder_JCh, _ax_scatter_Jpapbp
from scicomap.cblind import _get_color_weak_cmap, colorblind_vision
from scicomap.utils import _pyramid, _pyramid_zombie, _fn_with_roots, \
    _periodic_fn, _plot_examples, _plot_examples_qual, _complex_phase


class SciCoMap:
    """
    Get a matplotlib compatible color map from different packages.
    Mainly scientific color maps. Some are suited to a dark background and others for light backgrounds.

    Params:
    -------
    :param ctype: str
        color map type, either 'sequential', 'multi-sequential', 'diverging',
         'circular', 'miscellaneous'  or 'qualitative'
    :param cmap: str or cmap object
        the name of the color map you want to use or the matplotlib cmap object


    Attributes:
    -----------
    color_map_dic: dict
        the mapping dictionary for some colormaps
    ctype: str,
        color map type, either 'sequential', 'diverging' or 'qualitative'
    cname: str,
        the name of the color map you want to use
    cmap: matplotlib cmap or list of hex/rgb
        the color map
    uniformized: Boolean
        if the cmap has been uniformized or not

    Methods:
    --------
    get_ctype()
        get the colormap type

    get_mpl_color_map()
        get the matplotlib color map (cmap object)

    uniformize_cmap(lift=None)
        uniformize the colormap (linearize the brightness J')
        You can use `lift` to make it brighter (increase the
        minimum brightness of the colormap)

    symmetrize_cmap(bitonic=True, diffuse=True)
        symmetrise the hue (h'), using bi-tonic or not
        and smoothing or not the hue curve

    unif_sym_cmap(lift=None, bitonic=True, diffuse=True)
        uniformize and symmetrize at once (first uniformize and then symmetrize)

    get_color_map_names()
        get the name of all the available color maps

    get_color_map_dic()
        get the color maps dictionary

    assess_cmap(figsize=(18, 8))
        plot the Jch tensor to visualize the brightness (J'), the hue (h')
        and the chroma (c')

    illustrate_palettes(figsize=(12, 10), n_colors=256):
        plot the gradient or the discrete palettes of all the colormaps of the given type

    colorblind(figsize=(12, 5), n_colors=256, facecolor="black")
        plot the gradient or barchart of the colormap for different kinds of color deficiencies

    Example:
    --------
    ccmap = SciCoMap(ctype='sequential', cname='thermal')
    mpl_map = ccmap.get_mpl_color_map()



    References:
    -----------
    https://www.kennethmoreland.com/color-advice/
    https://mycarta.wordpress.com/2012/05/29/the-rainbow-is-dead-long-live-the-rainbow-series-outline/
    http://www.fabiocrameri.ch/colourmaps.php
    https://betterfigures.org/2015/06/23/picking-a-colour-scale-for-scientific-graphics/

    etc.
    """

    def __init__(self, ctype="sequential", cmap="thermal"):

        self.ctype = ctype
        self.color_map_dic = get_cmap_dict()

        self.cmap = cmap
        self.cname = 'cmap'
        self.uniformized = False

    def __repr__(self):

        s = "SciCoMap(ctype={ctype}, \n" \
            "     cmap={cmap})".format(ctype=self.ctype, cmap=self.cname)

        return s

    @classmethod
    def get_ctype(cls):
        """ return the colormap type """
        return get_cmap_dict().keys()

    def get_mpl_color_map(self):
        """
        Get the matplotlib colormap object

        :return:
         color_map: mpl colormap
        """
        if isinstance(self.cmap, str):
            if self.cmap not in self.color_map_dic[self.ctype].keys():
                raise ValueError(
                    "Current builtin cmap are: {}".format(
                        self.color_map_dic[self.ctype].keys()
                    )
                )
            self.cname = self.cmap
            self.cmap = self.color_map_dic[self.ctype][self.cmap]
        else:
            self.cname = self.cmap.name

        return self.cmap

    def uniformize_cmap(self, lift=None):
        """
        Uniformize the colormap, meaning linearizing the brightness (J')
        in the CAM02-UCS color space.

        :param lift: None or int in [0, 100]
            lift or not the darkest part of the cmap

        :return:
        """
        self.get_mpl_color_map()
        self.cmap, self.uniformized = uniformize_cmap(cmap=self.cmap,
                                                      name=self.cname,
                                                      lift=lift,
                                                      uniformized=self.uniformized)

    def symmetrize_cmap(self, bitonic=True, diffuse=True):
        """
        Symmetrize the colormap (the hue) in the CAM02-UCS color space.
        It can be symmetrized in a bitonic way or not (if bitonic, the hue
        curve will be symmetric with an extremum at its centre)

        The hue curve can be smoothed (diffuse) or not (edges might occur)

        :param bitonic: Boolean, default=True
            Bitonic symmetrization or not (extremum located at the centre of the hue curve)
        :param diffuse: Boolean, default=True
            Smooth hue curve or not (if not, edges might occur)
        :return:
        """
        self.get_mpl_color_map()
        self.cmap = symmetrize_cmap(cmap=self.cmap,
                                    name=self.cname,
                                    bitonic=bitonic,
                                    diffuse=diffuse)

    def unif_sym_cmap(self, lift=None, bitonic=True, diffuse=True):
        """
        First, uniformize the colormap, meaning linearizing the brightness (J')
        in the CAM02-UCS color space.

        Second, symmetrize the colormap (the hue) in the CAM02-UCS color space.
        It can be symmetrized in a bitonic way or not (if bitonic, the hue
        curve will be symmetric with an extremum at its centre)

        The hue curve can be smoothed (diffuse) or not (edges might occur)

        :param lift: None or int in [0, 100]
            lift or not the darkest part of the cmap
        :param bitonic: Boolean, default=True
            Bitonic symmetrization or not (extremum located at the centre of the hue curve)
        :param diffuse: Boolean, default=True
            Smooth hue curve or not (if not, edges might occur)
        :return:
        """
        self.get_mpl_color_map()
        self.cmap, self.uniformized = unif_sym_cmap(cmap=self.cmap,
                                                    name=self.cname,
                                                    lift=lift,
                                                    uniformized=self.uniformized,
                                                    bitonic=bitonic,
                                                    diffuse=diffuse)

    def get_color_map_names(self):
        """
        Get the names of the implemented colormaps for the chosen ctype

        :return: list
            list of colormap names
        """
        return self.color_map_dic[self.ctype].keys()

    @classmethod
    def get_color_map_dic(cls):
        """
        Get the mapping dict of all the available color maps

        :return: dict,
            the dictionary of all the color maps for all ctypes
        """
        return get_cmap_dict()

    def assess_cmap(self, figsize=(18, 8)):
        """
        Stolen from the ehtplot package

        Plot J', C', and h' of a colormap as function of the mapped value

        The CAM02-UCS lightness J' is linearized for
        generating perceptually uniform colormaps (working definition of Perceptually
        Uniform Sequential colormaps by matplotlib).

        Hue h' can encode an additional physical quantity in an image
        (when used in this way, the change of hue should be linearly
        proportional to the quantity)

        The other dimension chroma is less recognizable and should not be
        used to encode physical information. Since sRGB is only a subset
        of the Lab color space, there are human recognizable colors that
        are not displayable. In order to accurately represent the physical
        quantities

        :param figsize: 2-uple of int
            the figure size
        :return fig object
            the matplotlib figure object
        """
        color_map = self.get_mpl_color_map()
        return jch_plot(cmap=color_map, figsize=figsize)

    def illustrate_palettes(self, figsize=(12, 10), n_colors=256, facecolor="black"):
        """
        Draw the gradient or discrete color palettes for each colormaps of the chosen ctype

        :param figsize: 2-uple of int
            the figure size
        :param n_colors: int, default=10
            the number of colors to plot (e.g. 10 for qualitative and 256 for continuous)
        :param facecolor: str
            the chart face color. It should be a string of builtin matplotlib colors or a string
            corresponding to a hex color.

        :return:
        """
        cmap_list = self.get_color_map_names()
        ctype = self.ctype
        plot_colormap(ctype, cmap_list, figsize, n_colors, facecolor)
        plt.show()

    def colorblind(self, figsize=(12, 5), n_colors=256, facecolor="black"):
        """
        Draw the gradient or discrete color palettes for different kind of color vision deficiencies
        (provide the color vision deficiencies rendering of the visible spectrum for comparison).

        :param figsize: 2-uple of int
            the figure size
        :param n_colors: int, default=10
            the number of colors to plot (e.g. 10 for qualitative and 256 for continuous)
        :param facecolor: str
            the chart face color. It should be a string of builtin matplotlib colors or a string
            corresponding to a hex color.

        :return:
        """
        plot_colorblind_vision(ctype=self.ctype,
                               cmap_list=[self.get_mpl_color_map()],
                               figsize=figsize,
                               n_colors=n_colors,
                               facecolor=facecolor)


class ScicoSequential(SciCoMap):
    """
    Get a matplotlib compatible sequential color map from different packages
    providing scientific color maps. The color map can be uniformized and symmetrized if needed.
    The perception can be assessed as well as the color vision deficient rendering.

    This is useful for continuous values, for which there is no "centre" or mid value.

    Some are suited to a dark background and other for light backgrounds.

    Params:
    -------
    :param cmap: str or matplotlib cmap
        the name of the color map you want to use


    Attributes:
    -----------
    color_map_dic: dict
        the mapping dictionary for some colormaps
    ctype: str,
        color map type, either 'sequential', 'diverging' or 'qualitative'
    cname: str,
        the name of the color map you want to use
    cmap: matplotlib cmap or list of hex/rgb
        the color map

    Methods:
    --------
    get_mpl_color_map()
        get the matplotlib color map (or list of hex/rgb colors for qualitative)

    get_color_map_names()
        get the name of all the available color maps

    get_color_map_dic()
        get the color maps dictionary

    illustrate_palettes():
        plot the gradient or the discrete palettes (all of them)

    draw_example(facecolor="black")
        draw two charts for illustrative purposes

    Example:
    --------
    sc_map = ScicoSequential(cname='chroma')
    mpl_map = sc_map.get_mpl_color_map()
    sc_map.draw_example()



    References:
    -----------
    https://www.kennethmoreland.com/color-advice/
    https://mycarta.wordpress.com/2012/05/29/the-rainbow-is-dead-long-live-the-rainbow-series-outline/
    http://www.fabiocrameri.ch/colourmaps.php
    https://betterfigures.org/2015/06/23/picking-a-colour-scale-for-scientific-graphics/
    """

    def __init__(self, cmap="thermal"):
        super().__init__(cmap=cmap, ctype="sequential")

    def __repr__(self):

        s = "ScicoSequential(cmap={cmap})".format(cmap=self.cname)

        return s


    def draw_example(self, facecolor="black", figsize=(20, 20), cblind=True):
        color_map = self.get_mpl_color_map()
        elevation = load_hill_topography()
        scan_im = load_scan_image()
        xpyr, ypyr, zpyr = _pyramid()
        per_x, per_z, per_z = _periodic_fn()
        images = [elevation, scan_im, zpyr, "3D", per_z, "3D"]

        fig = _plot_examples(color_map=color_map,
                             images=images,
                             arr_3d=[(xpyr, ypyr, zpyr), (per_x, per_z, per_z)],
                             figsize=figsize,
                             facecolor=facecolor,
                             cname=self.cname,
                             cblind=cblind)

        return fig


class ScicoMultiSequential(SciCoMap):
    """
    Get a matplotlib compatible sequential color map from different packages.
    This is useful for continuous values, for which there is no "centre" or mid value
    (if there is one, use ScicoDiverging)
    Some are suited to a dark background and others for light backgrounds.

    Params:
    -------
    :param cmap: str or matplotlib cmap
        the name of the color map you want to use


    Attributes:
    -----------
    color_map_dic: dict
        the mapping dictionary for some colormaps
    ctype: str,
        color map type, either 'sequential', 'diverging' or 'qualitative'
    cname: str,
        the name of the color map you want to use
    cmap: matplotlib cmap or list of hex/rgb
        the color map

    Methods:
    --------
    get_mpl_color_map()
        get the matplotlib color map (or list of hex/rgb colors for qualitative)

    get_color_map_names()
        get the name of all the available color maps

    get_color_map_dic()
        get the color maps dictionary

    illustrate_palettes():
        plot the gradient or the discrete palettes (all of them)

    draw_example(facecolor="black")
        draw two charts for illustrative purposes

    Example:
    --------
    sc_map = ScicoSequential(cname='chroma')
    mpl_map = sc_map.get_mpl_color_map()
    sc_map.draw_example()



    References:
    -----------
    https://www.kennethmoreland.com/color-advice/
    https://mycarta.wordpress.com/2012/05/29/the-rainbow-is-dead-long-live-the-rainbow-series-outline/
    http://www.fabiocrameri.ch/colourmaps.php
    https://betterfigures.org/2015/06/23/picking-a-colour-scale-for-scientific-graphics/
    """

    def __init__(self, cmap="chroma"):
        super().__init__(cmap=cmap, ctype="multi-sequential")

    def __repr__(self):

        s = "ScicoMultiSequential(cmap={cmap})".format(cmap=self.cname)

        return s

    def draw_example(self, facecolor="black", figsize=(20, 20), cblind=True):
        color_map = self.get_mpl_color_map()
        elevation = load_hill_topography()
        scan_im = load_scan_image()
        xpyr, ypyr, zpyr = _pyramid_zombie(stacked=True)
        per_x, per_z, per_z = _periodic_fn()
        images = [elevation, scan_im, zpyr, "3D", per_z, "3D"]

        fig = _plot_examples(color_map=color_map,
                             images=images,
                             arr_3d=[(xpyr, ypyr, zpyr), (per_x, per_z, per_z)],
                             figsize=figsize,
                             facecolor=facecolor,
                             cname=self.cname,
                             cblind=cblind)

        return fig


class ScicoDiverging(SciCoMap):
    """
    Get a matplotlib compatible diverging color map from different packages.
    This is useful for continuous values, for which there is a "centre" or mid value
    (if there isn't, use ScicoSequential)
    Some are suited to a dark background and others for light backgrounds.

    Params:
    -------
    :param cmap: str or matplotlib cmap
        the name of the color map you want to use

    Attributes:
    -----------
    color_map_dic: dict
        the mapping dictionary for some colormaps
    ctype: str,
        color map type, either 'sequential', 'diverging' or 'qualitative'
    cname: str,
        the name of the color map you want to use
    cmap: matplotlib cmap or list of hex/rgb
        the color map

    Methods:
    --------
    get_mpl_color_map()
        get the matplotlib color map (or list of hex/rgb colors for qualitative)

    get_color_map_names()
        get the name of all the available color maps

    get_color_map_dic()
        get the color maps dictionary

    illustrate_palettes():
        plot the gradient or the discrete palettes (all of them)

    draw_example(facecolor="black")
        draw two charts for illustrative purposes

    Example:
    --------
    sc_map = ScicoDiverging(cname='redshift')
    mpl_map = sc_map.get_mpl_color_map()
    sc_map.draw_example()



    References:
    -----------
    https://www.kennethmoreland.com/color-advice/
    https://mycarta.wordpress.com/2012/05/29/the-rainbow-is-dead-long-live-the-rainbow-series-outline/
    http://www.fabiocrameri.ch/colourmaps.php
    https://betterfigures.org/2015/06/23/picking-a-colour-scale-for-scientific-graphics/
    """

    def __init__(self, cmap="wildfire"):
        super().__init__(cmap=cmap, ctype="diverging")

    def __repr__(self):

        s = "ScicoDiverging(cmap={cmap})".format(cmap=self.cname)

        return s

    def draw_example(self, facecolor="black", figsize=(20, 20)):
        color_map = self.get_mpl_color_map()
        # Create diverging image data
        image_div = _fn_with_roots()
        xpyr, ypyr, zpyr = _pyramid_zombie(stacked=False)
        per_x, per_z, per_z = _periodic_fn()

        images = [image_div, zpyr, '3D', per_z, "3D"]

        fig = _plot_examples(color_map=color_map,
                             images=images,
                             arr_3d=[(xpyr, ypyr, zpyr), (per_x, per_z, per_z)],
                             figsize=figsize,
                             facecolor=facecolor,
                             cname=self.cname,
                             norm=True)

        return fig


class ScicoCircular(SciCoMap):
    """
    Get a matplotlib compatible circular color map from different packages.
    This is useful for anglular values (circular "flat" as the phase cmap),
    for which there is no "centre" or mid value
    (if there is one, use ScicoDiverging)
    Some are suited to a dark background and others for light backgrounds.

    Params:
    -------
    :param cmap: str or matplotlib cmap
        the name of the color map you want to use


    Attributes:
    -----------
    color_map_dic: dict
        the mapping dictionary for some colormaps
    ctype: str,
        color map type, either 'sequential', 'diverging' or 'qualitative'
    cname: str,
        the name of the color map you want to use
    cmap: matplotlib cmap or list of hex/rgb
        the color map

    Methods:
    --------
    get_mpl_color_map()
        get the matplotlib color map (or list of hex/rgb colors for qualitative)

    get_color_map_names()
        get the name of all the available color maps

    get_color_map_dic()
        get the color maps dictionary

    illustrate_palettes():
        plot the gradient or the discrete palettes (all of them)

    draw_example(facecolor="black")
        draw two charts for illustrative purposes

    Example:
    --------
    sc_map = ScicoSequential(cname='chroma')
    mpl_map = sc_map.get_mpl_color_map()
    sc_map.draw_example()



    References:
    -----------
    https://www.kennethmoreland.com/color-advice/
    https://mycarta.wordpress.com/2012/05/29/the-rainbow-is-dead-long-live-the-rainbow-series-outline/
    http://www.fabiocrameri.ch/colourmaps.php
    https://betterfigures.org/2015/06/23/picking-a-colour-scale-for-scientific-graphics/
    """

    def __init__(self, cmap="colorwheel"):
        super().__init__(cmap=cmap, ctype="circular")

    def __repr__(self):

        s = "ScicoCircular(cmap={cmap})".format(cmap=self.cname)

        return s

    def draw_example(self, facecolor="black", figsize=(20, 20), cblind=True):
        color_map = self.get_mpl_color_map()
        elevation = load_hill_topography()
        scan_im = load_scan_image()
        per_x, per_z, per_z = _periodic_fn()
        images = [elevation, scan_im, "electric", "complex"]

        fig = _plot_examples(color_map=color_map,
                             images=images,
                             arr_3d=None,
                             figsize=figsize,
                             facecolor=facecolor,
                             cname=self.cname,
                             cblind=cblind)

        return fig


class ScicoMiscellaneous(SciCoMap):
    """
    Get a matplotlib compatible sequential color map from different packages.
    This is useful for continuous values, for which there is no "center" or mid value
    (if there is one, use ScicoDiverging)
    Some are suited to a dark background and others for light backgrounds.

    Params:
    -------
    :param cmap: str or matplotlib cmap
        the name of the color map you want to use


    Attributes:
    -----------
    color_map_dic: dict
        the mapping dictionary for some colormaps
    ctype: str,
        color map type, either 'sequential', 'diverging' or 'qualitative'
    cname: str,
        the name of the color map you want to use
    cmap: matplotlib cmap or list of hex/rgb
        the color map

    Methods:
    --------
    get_mpl_color_map()
        get the matplotlib color map (or list of hex/rgb colors for qualitative)

    get_color_map_names()
        get the name of all the available color maps

    get_color_map_dic()
        get the color maps dictionary

    illustrate_palettes():
        plot the gradient or the discrete palettes (all of them)

    draw_example(facecolor="black")
        draw two charts for illustrative purposes

    Example:
    --------
    sc_map = ScicoSequential(cname='chroma')
    mpl_map = sc_map.get_mpl_color_map()
    sc_map.draw_example()



    References:
    -----------
    https://www.kennethmoreland.com/color-advice/
    https://mycarta.wordpress.com/2012/05/29/the-rainbow-is-dead-long-live-the-rainbow-series-outline/
    http://www.fabiocrameri.ch/colourmaps.php
    https://betterfigures.org/2015/06/23/picking-a-colour-scale-for-scientific-graphics/
    """

    def __init__(self, cmap="turbo"):
        super().__init__(cmap=cmap, ctype="miscellaneous")

    def __repr__(self):

        s = "ScicoMiscellaneous(cmap={cmap})".format(cmap=self.cname)

        return s

    def draw_example(self, facecolor="black", figsize=(20, 20)):
        color_map = self.get_mpl_color_map()
        # Create diverging image data
        image_div = _fn_with_roots()
        xpyr, ypyr, zpyr = _pyramid_zombie(stacked=False)
        per_x, per_z, per_z = _periodic_fn()

        images = [image_div, zpyr, '3D', per_z, "3D"]

        fig = _plot_examples(color_map=color_map,
                             images=images,
                             arr_3d=[(xpyr, ypyr, zpyr), (per_x, per_z, per_z)],
                             figsize=figsize,
                             facecolor=facecolor,
                             cname=self.cname,
                             norm=True)

        return fig


class ScicoQualitative(SciCoMap):
    """
    Get a matplotlib compatible qualitative list of colors from different packages.
    This is useful for discrete values or categorical variables.
    Some are suited to a dark background and others for light backgrounds.

    Params:
    -------
    :param cmap: str or matplotlib cmap
        the name of the color map you want to use


    Attributes:
    -----------
    color_map_dic: dict
        the mapping dictionary for some colormaps
    ctype: str,
        color map type, either 'sequential', 'diverging' or 'qualitative'
    cname: str,
        the name of the color map you want to use
    cmap: matplotlib cmap or list of hex/rgb
        the color map

    Methods:
    --------
    get_mpl_color_map()
        get the matplotlib color map (or list of hex/rgb colors for qualitative)

    get_color_map_names()
        get the name of all the available color maps

    get_color_map_dic()
        get the color maps dictionary

    illustrate_palettes():
        plot the gradient or the discrete palettes (all of them)

    draw_example(facecolor="black")
        draw two charts for illustrative purposes

    Example:
    --------
    sc_map = ScicoQualitative(cname='glasbey_dark')
    mpl_map = sc_map.get_mpl_color_map()
    sc_map.draw_example()



    References:
    -----------
    https://www.kennethmoreland.com/color-advice/
    https://mycarta.wordpress.com/2012/05/29/the-rainbow-is-dead-long-live-the-rainbow-series-outline/
    http://www.fabiocrameri.ch/colourmaps.php
    https://betterfigures.org/2015/06/23/picking-a-colour-scale-for-scientific-graphics/
    """

    def __init__(self, cmap="glasbey_dark"):
        super().__init__(cmap=cmap, ctype="qualitative")

    def __repr__(self):

        s = "ScicoQualitative(cmap={cmap})".format(cmap=self.cname)

        return s

    def draw_example(self, facecolor="black", figsize=(20, 20)):
        color_map = self.get_mpl_color_map()

        # data from United Nations World Population Prospects (Revision 2019)
        # https://population.un.org/wpp/, license: CC BY 3.0 IGO
        year = [1950, 1960, 1970, 1980, 1990, 2000, 2010, 2018]
        population_by_continent = {
            'africa': [228, 284, 365, 477, 631, 814, 1044, 1275],
            'americas': [340, 425, 519, 619, 727, 840, 943, 1006],
            'asia': [1394, 1686, 2120, 2625, 3202, 3714, 4169, 4560],
            'europe': [220, 253, 276, 295, 310, 303, 294, 293],
            'oceania': [12, 15, 19, 22, 26, 31, 36, 39],
        }
        x = np.linspace(0, 10)
        # Fixing random state for reproducibility
        np.random.seed(19680801)
        noisy_trends = np.array(
            [np.sin(x) + x + np.random.randn(50),
             np.sin(x) + 0.5 * x + np.random.randn(50),
             np.sin(x) + 2 * x + np.random.randn(50),
             np.sin(x) - 0.5 * x + np.random.randn(50),
             np.sin(x) - 2 * x + np.random.randn(50),
             np.sin(x) + np.random.randn(50)]
        )
        noisy_trends = noisy_trends.T

        dict_arr = [population_by_continent, "scatter", noisy_trends]

        return _plot_examples_qual(color_map=color_map,
                                   dict_arr=dict_arr,
                                   figsize=figsize,
                                   facecolor=facecolor,
                                   cname=self.cname,
                                   year=year)


def get_cmap_dict():
    cmap_dict = {
        "diverging": {
            "berlin": scico.berlin,
            "bjy": cc.cm.bjy,
            "bky": cc.cm.bky,
            "BrBG": plt.get_cmap("BrBG"),
            "broc": scico.broc,
            "bwr": plt.get_cmap("bwr"),
            "coolwarm": plt.get_cmap("coolwarm"),
            "curl": cmocean.cm.curl,
            "delta": cmocean.cm.delta,
            "fusion": cmr.fusion,
            "fusion_r": cmr.fusion_r,
            "guppy": cmr.guppy,
            "guppy_r": cmr.guppy_r,
            "iceburn": cmr.iceburn,
            "iceburn_r": cmr.iceburn_r,
            "lisbon": scico.lisbon,
            "PRGn": plt.get_cmap("PRGn"),
            "PiYG": plt.get_cmap("PiYG"),
            "pride": cmr.pride,
            "pride_r": cmr.pride_r,
            "PuOr": plt.get_cmap("PuOr"),
            "RdBu": plt.get_cmap("RdBu"),
            "RdGy": plt.get_cmap("RdGy"),
            "RdYlBu": plt.get_cmap("RdYlBu"),
            "RdYlGn": plt.get_cmap("RdYlGn"),
            "redshift": cmr.redshift,
            "redshift_r": cmr.redshift_r,
            "roma": scico.roma,
            "seasons_r": cmr.seasons_r,
            "seismic": plt.get_cmap("seismic"),
            "spectral": plt.get_cmap("Spectral"),
            "turbo": plt.get_cmap("turbo"),
            "vanimo": scico.vanimo,
            "vik": scico.vik,
            "viola": cmr.viola,
            "viola_r": cmr.viola_r,
            "waterlily": cmr.waterlily,
            "waterlily_r": cmr.waterlily_r,
            "watermelon": cmr.watermelon,
            "watermelon_r": cmr.watermelon_r,
            "wildfire": cmr.wildfire,
            "wildfire_r": cmr.wildfire_r,
        },
        "sequential": {
            "afmhot": plt.get_cmap("afmhot"),
            "amber": cmr.amber,
            "amber_r": cmr.amber_r,
            "amp": cmocean.cm.amp,
            "apple": cmr.apple,
            "apple_r": cmr.apple_r,
            "autumn": plt.get_cmap("autumn"),
            "batlow": scico.batlow,
            "bilbao": scico.bilbao,
            "bilbao_r": scico.bilbao_r,
            "binary": plt.get_cmap("binary"),
            "Blues": plt.get_cmap("Blues"),
            "bone": plt.get_cmap("bone"),
            "BuGn": plt.get_cmap("BuGn"),
            "BuPu": plt.get_cmap("BuPu"),
            "chroma": cmr.chroma,
            "chroma_r": cmr.chroma_r,
            "cividis": plt.get_cmap("cividis"),
            "cool": plt.get_cmap("cool"),
            "copper": plt.get_cmap("copper"),
            "cosmic": cmr.cosmic,
            "cosmic_r": cmr.cosmic_r,
            "deep": cmocean.cm.deep,
            "dense": cmocean.cm.dense,
            "dusk": cmr.dusk,
            "dusk_r": cmr.dusk_r,
            "eclipse": cmr.eclipse,
            "eclipse_r": cmr.eclipse_r,
            "ember": cmr.ember,
            "ember_r": cmr.ember_r,
            "fall": cmr.fall,
            "fall_r": cmr.fall_r,
            "gem": cmr.gem,
            "gem_r": cmr.gem_r,
            "gist_gray": plt.get_cmap("gist_gray"),
            "gist_heat": plt.get_cmap("gist_heat"),
            "gist_yarg": plt.get_cmap("gist_yarg"),
            "GnBu": plt.get_cmap("GnBu"),
            "Greens": plt.get_cmap("Greens"),
            "gray": plt.get_cmap("gray"),
            "Greys": plt.get_cmap("Greys"),
            "haline": cmocean.cm.haline,
            "hawaii": scico.hawaii,
            "hawaii_r": scico.hawaii,
            "heat": cmr.torch,
            "heat_r": cmr.torch_r,
            "hot": plt.get_cmap("hot"),
            "ice": cmocean.cm.ice,
            "inferno": plt.get_cmap("inferno"),
            "imola": scico.imola,
            "imola_r": scico.imola_r,
            "lapaz": scico.lapaz,
            "lapaz_r": scico.lapaz_r,
            "magma": plt.get_cmap("magma"),
            "matter": cmocean.cm.matter,
            "neon": cmr.neon,
            "neon_r": cmr.neon_r,
            "neutral": cmr.neutral,
            "neutral_r": cmr.neutral_r,
            "nuuk": scico.nuuk,
            "nuuk_r": scico.nuuk,
            "ocean": cmr.ocean,
            "ocean_r": cmr.ocean_r,
            "OrRd": plt.get_cmap("OrRd"),
            "Oranges": plt.get_cmap("Oranges"),
            "pink": plt.get_cmap("pink"),
            "plasma": plt.get_cmap("plasma"),
            "PuBu": plt.get_cmap("PuBu"),
            "PuBuGn": plt.get_cmap("PuBuGn"),
            "PuRd": plt.get_cmap("PuRd"),
            "Purples": plt.get_cmap("Purples"),
            "rain": cmocean.cm.rain,
            "rainbow": perceptual_rainbow_16.mpl_colormap,
            "rainbow-sc": scico.batlow,
            "rainbow-sc_r": scico.batlow_r,
            "rainforest": cmr.rainforest,
            "rainforest_r": cmr.rainforest_r,
            "RdPu": plt.get_cmap("RdPu"),
            "Reds": plt.get_cmap("Reds"),
            "savanna": cmr.savanna,
            "savanna_r": cmr.savanna_r,
            "sepia": cmr.sepia,
            "sepia_r": cmr.sepia_r,
            "speed": cmocean.cm.speed,
            "solar": cmocean.cm.solar,
            "spring": plt.get_cmap("spring"),
            "summer": plt.get_cmap("summer"),
            "tempo": cmocean.cm.tempo,
            "thermal": cmocean.cm.thermal,
            "thermal_r": cmocean.cm.thermal_r,
            "thermal-2": cc.cm.bmy,
            "tokyo": scico.tokyo,
            "tokyo_r": scico.tokyo_r,
            "tropical": cmr.tropical,
            "tropical_r": cmr.tropical_r,
            "turbid": cmocean.cm.turbid,
            "turku": scico.turku,
            "turku_r": scico.turku_r,
            "viridis": plt.get_cmap("viridis"),
            "winter": plt.get_cmap("winter"),
            "Wistia": plt.get_cmap("Wistia"),
            "YlGn": plt.get_cmap("YlGn"),
            "YlGnBu": plt.get_cmap("YlGnBu"),
            "YlOrBr": plt.get_cmap("YlOrBr"),
            "YlOrRd": plt.get_cmap("YlOrRd"),
        },
        "multi-sequential": {
            "bukavu": scico.bukavu,
            "fes": scico.fes,
            "infinity": cmr.infinity,
            "infinity_s": cmr.infinity_s,
            "oleron": scico.oleron,
            "topo": cmocean.cm.topo,
        },
        "circular": {
            "bamo": scico.bamO,
            "broco": scico.brocO,
            "cet_c1": cc.cm.CET_C1,
            "colorwheel": cc.cm.colorwheel,
            "corko": scico.corkO,
            "phase": cmocean.cm.phase,
            "rainbow-iso": cc.cm.CET_I1,
            "romao": scico.romaO,
            "seasons": cmr.seasons,
            "seasons_s": cmr.seasons_s,
            "twilight": plt.get_cmap("twilight"),
            "twilight_s": plt.get_cmap("twilight_shifted"),
        },
        "miscellaneous": {
            "oxy": cmocean.cm.oxy,
            "rainbow-kov": cc.cm.rainbow,
            "turbo": plt.get_cmap("turbo"),
        },
        "qualitative": {
            "538": ListedColormap([[0, 143 / 255, 213 / 255],
                                   [252 / 255, 79 / 255, 48 / 255],
                                   [229 / 255, 174 / 255, 56 / 255],
                                   [109 / 255, 144 / 255, 79 / 255],
                                   [139 / 255, 139 / 255, 139 / 255],
                                   [129 / 255, 15 / 255, 124 / 255]], name="538"),
            "bold": ListedColormap(Bold_10.mpl_colors, name="bold"),
            "brewer": ListedColormap(Set1_9.mpl_colors, name="brewer"),
            "colorblind": ListedColormap(
                [[0.1, 0.1, 0.1],
                 [230 / 255, 159 / 255, 0],
                 [86 / 255, 180 / 255, 233 / 255],
                 [0, 158 / 255, 115 / 255],
                 [213 / 255, 94 / 255, 0],
                 [0, 114 / 255, 178 / 255]], name="colorblind"),
            "glasbey": cc.cm.glasbey,
            "glasbey_bw": cc.cm.glasbey_bw,
            "glasbey_category10": cc.cm.glasbey_category10,
            "glasbey_dark": cc.cm.glasbey_dark,
            "glasbey_hv": cc.cm.glasbey_hv,
            "glasbey_light": cc.cm.glasbey_light,
            "pastel": ListedColormap(Pastel_10.mpl_colors, name="pastel"),
            "prism": ListedColormap(Prism_10.mpl_colors, name="prism"),
            "vivid": ListedColormap(Vivid_10.mpl_colors, name="vivid"),
        },
    }
    return cmap_dict


def get_available_ctype():
    """ return available the colormap type """
    return get_cmap_dict.keys()


def plot_colormap(ctype, cmap_list='all', figsize=None, n_colors=10, facecolor="black",
                  uniformize=True, symmetrize=False, unif_kwargs=None, sym_kwargs=None
                  ):
    """
    Plot the gradient of the corresponding color palette (or bar plot if qualitative)

    :param ctype: str, default="sequential"
        the color map type
    :param cmap_list: list of string or 'all', default='all
        list of color map names to draw
    :param figsize: 2-uple of int
        the figure size
    :param n_colors: int, default=10
        the number of colors to plot (e.g. 10 for qualitative and 256 for continuous)
    :param facecolor: str
        the chart face color. It should be a string of builtin matplotlib colors or a string
        corresponding to a hex color.
    :param uniformize: Boolean, default=True
        uniformize or not the cmap before plotting
    :param symmetrize: Boolean, default=False
        symmetrize or not the cmap before plotting
    :param unif_kwargs: dict or None
        the kwargs for the uniformize_cmap method
    :param sym_kwargs: dict or None
        the kwargs for the symmetrize_cmap method
    :return:
    """
    if sym_kwargs is None:
        sym_kwargs = {}
    if unif_kwargs is None:
        unif_kwargs = {}

    gradient = np.linspace(0, 1, n_colors)
    gradient = np.vstack((gradient, gradient))

    if cmap_list == 'all':
        cmap_list = list(SciCoMap(ctype=ctype).get_color_map_names())

    nrows = len(cmap_list)

    if figsize is None:
        figsize = (10, 0.25 * nrows)

    fontcolor = "white" if facecolor == "black" else "black"
    font = {'color': fontcolor, 'size': 16}
    fig, axes = plt.subplots(nrows=nrows, figsize=figsize, facecolor=facecolor)
    fig.subplots_adjust(top=0.95, bottom=0.01, left=0.2, right=0.99)
    axes[0].set_title("Colormaps", fontdict=font)

    for ax, name in zip(axes, cmap_list):

        cmap = SciCoMap(ctype=ctype, cmap=name)

        if uniformize:
            cmap.uniformize_cmap(**unif_kwargs)
        if symmetrize:
            cmap.symmetrize_cmap(**sym_kwargs)

        cmap = cmap.get_mpl_color_map()

        if ctype == "qualitative":
            col_map = cmap(range(10)) if cmap.N > 10 else cmap(range(cmap.N))
            x = np.linspace(0, 1, len(col_map))
            ax.bar(x, np.ones_like(x), color=col_map, width=1 / (len(col_map) - 1))
        else:
            ax.imshow(gradient, aspect="auto", cmap=cmap)

        pos = list(ax.get_position().bounds)
        x_text = pos[0] - 0.01
        y_text = pos[1] + pos[3] / 2.0

        font = {'color': fontcolor, 'size': 12}
        fig.text(x_text, y_text, name, va="center", ha="right", fontdict=font)

    # Turn off *all* ticks & spines, not just the ones with colormaps.
    for ax in axes:
        ax.set_axis_off()
    return fig


def plot_colorblind_vision(ctype='sequential', cmap_list='all', figsize=None,
                           n_colors=10, facecolor="black", uniformize=True,
                           symmetrize=False, unif_kwargs=None, sym_kwargs=None):
    """

    Render the color map (adjusted or not) in different color deficiencies vision


    :param ctype: str, default="sequential"
        the color map type
    :param cmap_list: list of string or 'all', default='all
        list of color map names to draw
    :param figsize: 2-uple of int
        the figure size
    :param n_colors: int, default=10
        the number of colors to plot (e.g. 10 for qualitative and 256 for continuous)
    :param facecolor: str
        the chart face color. It should be a string of builtin matplotlib colors or a string
        corresponding to a hex color.
    :param uniformize: Boolean, default=True
        uniformize or not the cmap before plotting
    :param symmetrize: Boolean, default=False
        symmetrize or not the cmap before plotting
    :param unif_kwargs: dict or None
        the kwargs for the uniformize_cmap method
    :param sym_kwargs: dict or None
        the kwargs for the symmetrize_cmap method
    :return:
    """
    if sym_kwargs is None:
        sym_kwargs = {}
    if unif_kwargs is None:
        unif_kwargs = {}
    cm_list = []

    if cmap_list == 'all':
        cmap_list = list(SciCoMap(ctype=ctype).get_color_map_names())

    for name in cmap_list:
        cmap = SciCoMap(ctype=ctype, cmap=name)
        if uniformize:
            cmap.uniformize_cmap(**unif_kwargs)
        if symmetrize:
            cmap.symmetrize_cmap(**sym_kwargs)
        cmap = cmap.get_mpl_color_map()
        cm_list.append(cmap)

    return colorblind_vision(cmap=cm_list, figsize=figsize, n_colors=n_colors, facecolor=facecolor)


def compare_cmap(image='scan', ctype="sequential", cm_list=None, ncols=3, uniformize=True,
                 symmetrize=False, unif_kwargs=None, sym_kwargs=None, facecolor="black"):
    """
    Utility function to visualize how the different color maps render the details and the information.
    You can pass the image of your choice, like a topographic profile for sequential and sea-earth level for
    diverging (negative and positive values) for instance.
    Some are suited to a dark background and others for light backgrounds.

    :param image: str or None
        the path to the jpg or png picture you want to use for the comparison or one
        of the builtin images as a pyramid image to visualize if there is any artifact
    :param ctype: str,
        either "sequential", "diverging", "qualitative"
    :param cm_list: list of str or None
        the list of cmaps you want to compare, if None all of them (for the chosen ctype) will be compared
    :param ncols: int
        the number of columns in the matplotlib subplot figure
    :param uniformize: Boolean, default=True
        uniformize or not the cmap before plotting
    :param symmetrize: Boolean, default=False
        symmetrize or not the cmap before plotting
    :param unif_kwargs: dict or None
        the kwargs for the uniformize_cmap method
    :param sym_kwargs: dict or None
        the kwargs for the symmetrize_cmap method
    :param facecolor: str
        the chart face color. It should be a string of builtin matplotlib colors or a string
        corresponding to a hex color.

    :return: f, matplotlib figure
    """

    if unif_kwargs is None:
        unif_kwargs = {}
    if sym_kwargs is None:
        sym_kwargs = {}

    if image is not None and not isinstance(image, str):
        raise TypeError("image should be a string or a path to an existing file")

    if (image is not None) and (image.endswith(("jpg", "jpeg", "png"))):
        img = mpimg.imread(image)
        lum_img = img[:, :, 0]
    elif image == "pyramid":
        lum_img = _pyramid()
    elif image == "topography":
        lum_img = load_hill_topography()
    elif image == "fn_roots":
        lum_img = _fn_with_roots()
    elif image == "scan":
        lum_img = load_scan_image()
    elif image == "phase":
        lum_img = _complex_phase()
    elif image == "grmhd":
        lum_img = load_pic(name=image)
    elif image == "vortex":
        lum_img = load_pic(name=image)
    elif image == "tng":
        lum_img = load_pic(name=image)
    else:
        lum_img = _pyramid()

    if cm_list is None:
        cm_list = list(SciCoMap(ctype=ctype).get_color_map_names())

    nrows = int(np.ceil(len(cm_list) / ncols))
    # delete non-used axes
    n_charts = len(cm_list)
    n_subplots = nrows * ncols

    figsize = (2 * ncols, 2.5 * nrows)
    f, axs = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        figsize=figsize,
        # subplot_kw={"aspect": 1},
        facecolor=facecolor,
        # gridspec_kw={"hspace": 0.0, "wspace": 0.5},
    )
    # Make the axes accessible with single indexing
    axs = axs.flatten()
    fontcolor = "white" if facecolor == "black" else "black"

    # loop over the columns to illustrate
    for i, color_map in enumerate(cm_list):
        # select the axis where the map will go
        if n_charts > 1:
            ax = axs[i]
        else:
            ax = axs

        chartcm = SciCoMap(ctype=ctype, cmap=color_map)

        if uniformize:
            chartcm.uniformize_cmap(**unif_kwargs)
        if symmetrize:
            chartcm.symmetrize_cmap(**sym_kwargs)

        ax.imshow(lum_img, cmap=chartcm.get_mpl_color_map())
        ax.set_title(color_map, fontsize=16, color=fontcolor)
        # Remove axis clutter
        ax.set_axis_off()

    if n_subplots > n_charts > 1:
        for i in range(n_charts, n_subplots):
            ax = axs[i]
            ax.set_axis_off()

    # Display the figure
    # plt.tight_layout(pad=0, w_pad=0.5, h_pad=0)
    f.subplots_adjust(wspace=0, hspace=.25)
    return f


def jch_plot(cmap, figsize=(12, 10)):
    """
    Stolen from the ehtplot package

    Plot J', C', and h' of a colormap as function of the mapped value

    The CAM02-UCS lightness J' should serve us as a good approximation for
    generating perceptually uniform colormaps. In fact, linearity in J'
    is used as the working definition of Perceptually Uniform Sequential
    colormaps by matplotlib.

    Hue h' can encode an additional physical quantity in an image
    (when used in this way, the change of hue should be linearly
    proportional to the quantity)

    The other dimension chroma is less recognizable and should not be
    used to encode physical information. Since sRGB is only a subset
    of the Lab color space, there are human recognizable colors that
    are not displayable. In order to accurately represent the physical
    quantities.

    :param cmap: string or matplotlib.colors.Colormap): The colormap to
            be plotted.
    :param figsize: 2-uple of int
        the figure size
    """
    f = plt.figure(figsize=figsize)
    c_maps, _ = _get_color_weak_cmap(cmap, n_images=2)
    color_map, deuter50_cm, prot50_cm, deuter100_cm, trit100_cm = c_maps

    title_str = cmap if isinstance(cmap, str) else cmap.name
    f.suptitle(title_str, fontsize=24)

    ax0 = f.add_subplot(2, 4, 1)
    ax0 = _ax_cylinder_JCh(ax0, cmap, title="Normal (90-95% of pop)")

    ax2 = f.add_subplot(2, 4, 3)
    ax2 = _ax_cylinder_JCh(ax2, deuter50_cm, title="Deuter-50%, RG-weak")

    ax4 = f.add_subplot(2, 4, 7)
    ax4 = _ax_cylinder_JCh(ax4, deuter100_cm, title="Deuter-100%, RG-blind")

    ax6 = f.add_subplot(2, 4, 5)
    ax6 = _ax_cylinder_JCh(ax6, trit100_cm, title="Trit-100%, BY deficient")

    ax3d = f.add_subplot(2, 4, 2, projection="3d", elev=25, azim=-75)
    ax3d = _ax_scatter_Jpapbp(ax3d, cmap, title="Normal (90-95% of pop)")

    ax3d2 = f.add_subplot(2, 4, 4, projection="3d", elev=25, azim=-75)
    ax3d2 = _ax_scatter_Jpapbp(ax3d2, deuter50_cm, title="Deuter-50%, RG-weak")

    ax3d3 = f.add_subplot(2, 4, 8, projection="3d", elev=25, azim=-75)
    ax3d3 = _ax_scatter_Jpapbp(ax3d3, deuter100_cm, title="Deuter-100%, RG-blind")

    ax3d4 = f.add_subplot(2, 4, 6, projection="3d", elev=25, azim=-75)
    ax3d4 = _ax_scatter_Jpapbp(ax3d4, trit100_cm, title="Trit-100%, BY deficient")

    plt.tight_layout()

    return f



"""
Module for performing color math. Heavily based on colorspacious, viscm and ethplot.
The main model of perceptual distance is the "CAM02-UCS" color-space
(Uniform Colour Space version of the CIECAM02).

This module uses Cartesian Lab and CIECAM02 color spaces and cylindrical
CIELCh (hereafter LCh) and CIEJCh (hereafter JCh) color spaces which have coordinates L*, J*, C*, and h.
The lightness coordinates L* and J* are identical to Lab and Jab. The chroma (relative saturation)
C* and hue h (in degree h°) are simply C* = sqrt(a*^2 + b*^2) and h = atan2(b*, a*) according
to Redness-Greenness a and Yellowness-Blueness b in their own coordinates.

https://github.com/liamedeiros/ehtplot/blob/7a0567496ba9ab72f4a541d5994352bbe4eac764/ehtplot/color/cmath.py
"""

import numpy as np
import warnings
from colorspacious import cspace_convert
from matplotlib.colors import Colormap, ListedColormap
from scipy.interpolate import CubicSpline


def get_ctab(cmap):
    """
    Get the tabular version of the colormap, with the same number of rows than colors
    :param cmap: matplotlib colormap
        the matplotlib colormap object to convert to a np.array
    :return: np.array
        the numpy array of the the colors
    """
    if isinstance(cmap, Colormap):
        return np.array([cmap(v) for v in np.linspace(0, 1, cmap.N)])
    elif isinstance(cmap, list):
        return np.array(cmap)
    else:
        TypeError("`cmap` is neither a matplotlib Colormap nor a list of str/uples")


def max_chroma(
        Jp, hp, Cpmin=0.0, Cpmax="auto", eps=1024 * np.finfo(np.float).eps, clip=True
):
    """
    Compute the maximum allowed chroma given lightness J' and hue h', from the CAM02-UCS
    color space.

    :param Jp: np.array
        The lightness (in the JCh space)
    :param hp: np.array
        The hue
    :param Cpmin: float, default=0.0
        The minimal chroma value
    :param Cpmax: float or 'auto'
        The maximal chroma value
    :param eps: float
        the precision
    :param clip: Boolean, default=True
        To clip or not the minimal and maximal values of the lightness
    :return:
    """
    Jpmin = 5.54015251457561e-22
    Jpmaxv = 98.98016
    Jpmax = 99.99871678107648

    if clip:
        Jp = np.clip(Jp, Jpmin, min(Jpmaxv, Jpmax))

    if np.any(Jp < Jpmin) or np.any(Jp > Jpmax):
        raise ValueError("J' out of range.")

    if np.any(Jp > Jpmaxv):
        raise ValueError(
            "J' is out of range such that the corresponding sRGB colorspace "
            + "is offset and C' == 0 is no longer a valid assumption."
        )

    if Cpmax == "auto":
        Cpmax = np.clip(np.sqrt(100 * Jp), 0, 64)

    CpU = np.full(len(Jp), Cpmax)  # np.full() works for both scalar and array
    CpL = np.full(len(Jp), Cpmin)  # np.full() works for both scalar and array

    for i in range(64):
        Cp = 0.5 * (CpU + CpL)

        # Fix when we hit machine precision
        need_fix = Cp == CpU
        Cp[need_fix] = CpL[need_fix]

        Jpapbp = np.stack([Jp, Cp * np.cos(hp), Cp * np.sin(hp)], axis=-1)
        sRGB = transform(Jpapbp, inverse=True)
        edge = 2.0 * np.amax(abs(sRGB - 0.5), -1)

        if 1.0 - eps <= np.min(edge) and np.max(edge) <= 1.0:
            break

        I = edge >= 1.0
        CpU[I] = Cp[I]
        CpL[~I] = Cp[~I]
    else:
        raise ArithmeticError("WARNING: max_chroma() has not fully converged")

    return Cp


def transform(ctab, src="sRGB1", dst="CAM02-UCS", inverse=False):
    """Transform a colortable between color spaces"""

    out = ctab.copy()
    if not inverse:
        out[:, :3] = cspace_convert(out[:, :3], src, dst)
    else:
        out[:, :3] = cspace_convert(out[:, :3], dst, src)
    return out


def interp(x, xp, yp):
    """Improve numpy's interp() function to allow decreasing `xp`"""
    if xp[0] < xp[-1]:
        return np.interp(x, xp, yp)
    else:
        return np.interp(x, np.flip(xp, 0), np.flip(yp, 0))


def extrema(a):
    """Find extrema in an array"""
    da = a[1:] - a[:-1]
    xa = da[1:] * da[:-1]
    return np.argwhere(xa <= 0.0)[:, 0] + 1


def classify(Jpapbp):
    """
    Classify a colormap as sequential, divergent, circular, multi-sequential
    """

    N = Jpapbp.shape[0]
    large_diff_lum = Jpapbp[:, 0].max() - Jpapbp[:, 0].min() > 0.5
    first_and_last_simeq = sum(abs(np.floor(Jpapbp[0, ...]) - np.floor(Jpapbp[-1, ...]))) < 4

    # some of the cmcrameri colormaps have kind of a small hook at the end of the array
    # artifact increasing the number of extrema by +1. A rough fix is not considering the
    # beginning  and the end of the cmap
    x = extrema(Jpapbp[10:-10, 0])
    x += 10

    if (len(x) == 1 or large_diff_lum) and first_and_last_simeq:
        return "circular-div"
    elif len(x) != 1 and first_and_last_simeq:
        return "circular-flat"
    elif len(x) == 0:
        return "sequential"
    # strictly, the extremum should be at the middle element set([(N + 1) // 2 - 1, N // 2]):
    # cmcrameri uses another parametrization and is not perfectly symmetric
    # in the Jc'h space
    elif len(x) == 1 and x[0] in np.arange((N + 1) // 2 - 5, N // 2 + 5, 1):
        return "divergent"
    elif len(x) == 1 and x[0] not in np.arange((N + 1) // 2 - 5, N // 2 + 5, 1):
        return "asym_div"
    elif len(x) == 2 and (x[1] == x[0] + 1):
        return "multiseq"
    else:
        return "unknown"


def uniformize(Jpapbp, JpL=None, JpR=None, Jplower=None, Jpupper=None):
    """Make a colormap uniform in lightness J' (linearizing)"""
    if JpL is None:
        JpL = Jpapbp[0, 0]
    if JpR is None:
        JpR = Jpapbp[-1, 0]

    if Jplower is not None:
        JpL, JpR = max(JpL, Jplower), max(JpR, Jplower)
    if Jpupper is not None:
        JpL, JpR = min(JpL, Jpupper), min(JpR, Jpupper)

    out = Jpapbp.copy()
    out[:, 0] = np.linspace(JpL, JpR, out.shape[0])
    out[:, 1] = interp(out[:, 0], Jpapbp[:, 0], Jpapbp[:, 1])
    out[:, 2] = interp(out[:, 0], Jpapbp[:, 0], Jpapbp[:, 2])
    return out


def factor(
        Cp, softening=1.0, bitonic=True, diffuse=True, CpL=None, CpR=None, verbose=False, diverging=False
):
    """Compute the factor required to perform several chroma operations"""
    S = Cp + softening
    s = S.copy()

    N = len(Cp)
    H = N // 2
    m = np.minimum(s[:H], np.flip(s[-H:]))

    # Bitonic only for non diverging
    if bitonic and (not diverging):  # force half of Cp increase monotonically
        if m[H - 1] > s[H]:
            m[H - 1] = s[H]
            if verbose:
                print("Enforce bitonic at {}".format(s[H]))
        for i in range(H - 1, 0, -1):
            if m[i - 1] > m[i]:
                m[i - 1] = m[i]
                if verbose:
                    print("Enforce bitonic at {}".format(m[i]))

    s[:+H] = m
    s[-H:] = np.flip(m)

    if CpL is not None:
        s[0] = CpL + softening
    if CpR is not None:
        s[-1] = CpR + softening

    if diffuse:  # diffuse s using forward Euler
        for i in range(N):
            s[1:-1] += 0.5 * (s[2:] + s[:-2] - 2.0 * s[1:-1])

    return s / S


def symmetrize(Jpapbp, **kwargs):
    """Make a sequential colormap symmetric in chroma C'"""
    out = Jpapbp.copy()
    Jp = out[:, 0]
    Cp = np.sqrt(out[:, 1] * out[:, 1] + out[:, 2] * out[:, 2])

    f = factor(Cp, **kwargs)
    out[:, 1] *= f
    out[:, 2] *= f
    return out


def adjust_sequential(Jpapbp, roundup=None, bi_seq=False):
    """API for uniformizing a sequential colormap"""

    if bi_seq:
        x_boundary = extrema(Jpapbp[:, 0])[0]
        Jpapbp1 = Jpapbp[:x_boundary + 1, ...].copy()
        Jpapbp2 = Jpapbp[x_boundary + 1:, ...].copy()
        Jp = Jpapbp1[:, 0]
        Jplower = min(Jp[0], Jp[-1])
        if roundup is not None:
            Jplower = np.ceil(Jplower / roundup) * roundup
        Jpapbp1 = uniformize(Jpapbp1, Jplower=Jplower)
        Jp = Jpapbp2[:, 0]
        Jplower = min(Jp[0], Jp[-1])
        if roundup is not None:
            Jplower = np.ceil(Jplower / roundup) * roundup
        Jpapbp2 = uniformize(Jpapbp2, Jplower=Jplower)
        return np.append(Jpapbp1, Jpapbp2, axis=0)
    else:
        Jp = Jpapbp[:, 0]
        Jplower = min(Jp[0], Jp[-1])
        if roundup is not None:
            Jplower = np.ceil(Jplower / roundup) * roundup

        return uniformize(Jpapbp, Jplower=Jplower)


def adjust_circular_flat(Jpapbp):
    """Adjusting the flat circular colormap, making the lightness constant"""
    out = Jpapbp.copy()
    Jp = out[:, 0].ravel()
    out[:, 0] = np.ones_like(Jp) * np.nanmean(Jp)
    return out


def adjust_circular(Jpapbp, roundup=None):
    """API for uniformizing a circular colormap (non flat in lightness)"""
    Jp = Jpapbp[:, 0]
    x_extr = extrema(Jp)
    h = x_extr[0]
    H = h + 1
    N = Jpapbp.shape[0]
    # h = (N + 1) // 2 - 1  # == H-1 if even; == H if odd
    # H = N // 2

    if Jp[1] > Jp[0]:  # hill
        Jplower = max(Jp[0], Jp[-1])
        Jpupper = min(Jp[h], Jp[H])
    else:  # valley
        Jplower = max(Jp[h], Jp[H])
        Jpupper = min(Jp[0], Jp[-1])
    if roundup is not None:
        Jplower = np.ceil(Jplower / roundup) * roundup

    L = uniformize(Jpapbp[: h + 1, :], Jplower=Jplower, Jpupper=Jpupper)
    R = uniformize(Jpapbp[H:, :], Jplower=Jplower, Jpupper=Jpupper)
    return np.append(L, R[N % 2:, :], axis=0)


def adjust_divergent(Jpapbp, roundup=None, circular=False, symmetric=True):
    """API for uniformizing a divergent colormap"""
    Jp = Jpapbp[:, 0]
    out = Jpapbp.copy()
    x_extr = extrema(Jp)
    h = x_extr[0]
    H = h + 1
    N = Jpapbp.shape[0]
    # h = (N + 1) // 2 - 1  # == H-1 if even; == H if odd
    # H = N // 2

    if Jp[1] > Jp[0] and symmetric:  # hill
        Jplower = max(Jp[0], Jp[-1])
        Jpupper = min(Jp[h], Jp[H])
    elif Jp[1] > Jp[0] and not symmetric:  # hill
        Jplower = Jp[0]
        Jpupper = Jp[h]
    elif Jp[1] < Jp[0] and symmetric:  # valley
        Jplower = max(Jp[h], Jp[H])
        Jpupper = min(Jp[0], Jp[-1])
    else:
        Jplower = Jp[h]
        Jpupper = Jp[-1]
    if roundup is not None:
        Jplower = np.ceil(Jplower / roundup) * roundup

    L = uniformize(Jpapbp[: h + 1, :], Jplower=Jplower, Jpupper=Jpupper)
    R = uniformize(Jpapbp[H:, :], Jplower=Jplower, Jpupper=Jpupper)
    Jpapbp_unif = np.append(L, R[N % 2:, :], axis=0)

    if circular:
        theta = 2 * np.pi * np.linspace(0, 1, out.shape[0])
        out[0, 1:3] = out[-1, 1:3]
        cs = CubicSpline(theta, out[:, 1:3], bc_type='periodic')
        Jpapbp_unif[:, 1:3] = cs(theta)
    return Jpapbp_unif


def uniformize_cmap(cmap, name="new_cmap", lift=None, uniformized=False):
    """
    Uniformize the color map (perceptually homogeneous)

    :param cmap: matplotlib cmap
        the cmap object to linearize
    :param name: str, default="new_cmap"
        the name of the new (created) cmap
    :param lift, None or int (0,100)
        whether or not increase the lightness (Jab parametrization)
    :param uniformized, Boolean
        if the cmap has been uniformized already, avoid redundant lift
    :return: cmap object
        a matplotlib colormap
    """
    # if not uniformized yet, uniformize the cmap
    # else do nothing
    if not uniformized:
        # get the color table
        ctab = get_ctab(cmap)
        # transform to Jpapbp
        t_ctab = transform(ctab)
        # find if sequential or divergent or unknown
        cmap_type = classify(t_ctab)
        if cmap_type == "divergent":
            lin_ctab = adjust_divergent(t_ctab, roundup=lift)
        elif cmap_type == "asym_div":
            lin_ctab = adjust_divergent(t_ctab, roundup=lift, symmetric=False)
        elif cmap_type == "sequential":
            lin_ctab = adjust_sequential(t_ctab, roundup=lift)
        elif cmap_type == "multiseq":
            lin_ctab = adjust_sequential(t_ctab, roundup=lift, bi_seq=True)
        elif cmap_type == "circular-flat":
            lin_ctab = adjust_circular_flat(t_ctab)
        elif cmap_type == "circular-div":
            lin_ctab = adjust_divergent(t_ctab, roundup=lift, circular=True)
        else:
            warnings.warn("The colormap {} type is unknown (not recognized as sequential or divergent)\n"
                          "Not uniformized".format(name))
            lin_ctab = t_ctab

        lin_cmap = transform(ctab=lin_ctab, inverse=True)

        return ListedColormap(np.clip(lin_cmap, 0, 1), name=name), True
    else:
        return cmap, True


def symmetrize_cmap(cmap, name="new_cmap", bitonic=True, diffuse=True):
    """
    Symmetrize the color map (symmetric saturation for even perception of both sides of the cmap)

    :param cmap: matplotlib cmap
        the cmap object to linearize
    :param name: str, default="new_cmap"
        the name of the new (created) cmap
    :param bitonic: Boolean, default=True
        if True, the chroma will be forced to increase monotonically,
        reach max or plateau and then decrease
        (bitonic, not suited for other than sequential cmaps).
    :param diffuse: Boolean, default=True
        diffuse colors, avoid sharp edges in Chroma (saturation)
    :return: cmap object
        a matplotlib colormap
    """
    # get the color table
    ctab = get_ctab(cmap)
    # transform to Jpapbp
    t_ctab = transform(ctab)
    # find if sequential or divergent or unknown
    cmap_type = classify(t_ctab)
    diverging = True if cmap_type == "divergent" else False
    lin_ctab = symmetrize(t_ctab, bitonic=bitonic, diffuse=diffuse, diverging=diverging)
    # get back a matplotlib cmap object
    s_cmap = transform(ctab=lin_ctab, inverse=True)
    return ListedColormap(np.clip(s_cmap, 0, 1), name=name)


def unif_sym_cmap(cmap, name="new_cmap", lift=None, uniformized=False, bitonic=True, diffuse=True):
    """
    Uniformize the color map (perceptually homogeneous) and Symmetrize the color map
    (symmetric saturation for even perception of both sides of the cmap)

    :param cmap: matplotlib cmap
        the cmap object to linearize
    :param name: str, default="new_cmap"
        the name of the new (created) cmap
    :param lift, None or int (0,100)
        whether or not increase the lightness (Jab parametrization)
    :param uniformized, Boolean
        if the cmap has been uniformized already, avoid redundant lift
    :param bitonic: Boolean, default=True
        if True, the chroma will be forced to increase monotonically,
        reach max or plateau and then decrease
        (bitonic, not suited for other than sequential cmaps).
    :param diffuse: Boolean, default=True
        diffuse colors, avoid sharp edges in Chroma (saturation)
    :return: cmap object
        a matplotlib colormap
    """
    uni_cmap, uniformized = uniformize_cmap(cmap, name=name, lift=lift, uniformized=uniformized)
    return symmetrize_cmap(uni_cmap, name=name, bitonic=bitonic, diffuse=diffuse), uniformized


def _ax_cylinder_JCh(ax, cmap, title):
    """Create the matplotlib ax for the JCh"""
    ctab = get_ctab(cmap)  # get the colormap as a color table in sRGB
    Jpapbp = transform(ctab)  # transform color table into CAM02-UCS colorspace

    Jp = Jpapbp[:, 0]
    ap = Jpapbp[:, 1]
    bp = Jpapbp[:, 2]

    Cp = np.sqrt(ap * ap + bp * bp)
    hp = np.arctan2(bp, ap) * 180 / np.pi
    v = np.linspace(0.0, 1.0, len(Jp))

    ax.set_title(title + " - JpCphp coord. (cyl)", fontsize=14)
    ax.set_xlabel("Value", fontsize=14)

    axtx = ax.twinx()
    ax.set_ylim(0, 100)
    ax.set_ylabel("J' & C' (0-100)", fontsize=14)
    axtx.set_ylim(-180, 180)
    axtx.set_ylabel("h' (degrees)", fontsize=14)

    ax.scatter(v, Jp, color=ctab, label="J' (lightness)")
    ax.plot(v, Cp, c="k", linestyle="--", label="C' (chroma)")
    axtx.scatter(v[::15], hp[::15], s=10, c="k")
    ax.scatter([], [], s=10, c="k", label="h' (hue)")
    ax.legend(loc="best")
    return ax


def _ax_scatter_Jpapbp(ax, cmap, title):
    """Create the matplotlib ax for the scatter plot in 3D"""
    ctab = get_ctab(cmap)  # get the colormap as a color table in sRGB
    Jpapbp = transform(ctab)  # transform color table into CAM02-UCS colorspace

    # ax.plot(Jpapbp[:, 1], Jpapbp[:, 2], Jpapbp[:, 0])
    x_dots = np.linspace(0, 1, Jpapbp.shape[0])
    RGB_dots = cmap(x_dots)[:, :3]
    Jpapbp_dots = transform(RGB_dots)
    ax.scatter(Jpapbp_dots[:, 1],
               Jpapbp_dots[:, 2],
               Jpapbp_dots[:, 0],
               c=RGB_dots[:, :],
               s=80)
    ax.set_xlabel("a' (green -> red)")
    ax.set_ylabel("b' (blue -> yellow)")
    ax.set_zlabel("J'/K (black -> white)")
    ax.set_title(title + " - Jpapbp coord.", fontsize=14)
    return ax



import itertools
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
from matplotlib.colors import ListedColormap
from matplotlib.patches import Circle
from mpl_toolkits.axes_grid1 import make_axes_locatable
from os.path import dirname, join
# internal import
from scicomap.cblind import _get_color_weak_ctab, _get_color_weak_cmap



def _pyramid(n=513):
    """Create a pyramid function"""
    s = np.linspace(-1.0, 1.0, n)
    x, y = np.meshgrid(s, s)
    z = 1.0 - np.maximum(abs(x), abs(y))
    return x, y, z


def _pyramid_zombie(n=513, stacked=False):
    """Create a pyramid and its zombie copy (buried pyramid)"""
    # image plot
    s_y = np.linspace(-1.0, 1.0, n)
    x, y = np.meshgrid(s_y, s_y)
    z = 1.0 - np.maximum(abs(x), abs(y))
    z = np.hstack((z, z - 1)) if stacked else np.hstack((z, -z))
    # 3d plot
    s_x = np.linspace(-1.0, 3.0, 2 * n)
    s_y = np.linspace(-1.0, 1.0, n)
    x, y = np.meshgrid(s_x, s_y)
    return x, y, z


def _periodic_fn():
    """Create a periodic function with a step function"""
    dx = dy = 0.05
    y, x = np.mgrid[-5: 5 + dy: dy, -5: 10 + dx: dx]
    z = np.sin(x) ** 10 + np.cos(10 + y * x) + np.cos(x) + 0.2 * y + 0.1 * x + \
        np.heaviside(x, 1) * np.heaviside(y, 1)
    z = z - np.mean(z)
    return x, y, z


def _fn_with_roots(n=250):
    """Create a function with roots, for diverging colormaps """
    x, y = np.meshgrid(np.linspace(-3, 3, n), np.linspace(-2, 2, n))
    return (1 - x / 2 + x ** 5 + y ** 3) * np.exp(-(x ** 2) - y ** 2)


def _complex_phase(n=100):
    """argument of a function in the complex plane, cyclic colormaps"""
    x, y = np.ogrid[-3:3:n * 1j, -3:3:n * 1j]
    z = x + 1j * y
    # w = (z ** 2 - 1) * (z - 2 - 1j) ** 2 / (z ** 2 + 2 + 2j)
    w = (z ** 2 - 1) * (z - 2 - 1j) ** 2
    return np.angle(w)


def _plot_complex_arg(ax, cmap, facecolor="black", title=False):
    """Create mlp ax of the complex argument"""
    arg_z = _complex_phase(n=100)
    title_color = "white" if facecolor == "black" else "black"
    ang = ax.imshow(np.degrees(arg_z), extent=[-2, 2, -2, 2], cmap=cmap)

    ax.set_xlim(-2, 2)
    ax.set_ylim(-2, 2)
    ax.set_aspect('equal')
    if title:
        ax.set_xlabel('Re', color=title_color, fontsize=12)
        ax.set_ylabel('Im', color=title_color, fontsize=12)
        ax.set_title(r"Argument of $(z^2 - 1)(z - 2 - j)^2$", color=title_color, fontsize=14)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cax.yaxis.label.set_color(title_color)
    cax.tick_params(axis='y', colors=title_color)
    plt.colorbar(ang, cax=cax)
    return ax


def _E(q, r0, x, y):
    """Return the electric field vector E=(Ex,Ey) due to charge q at r0."""
    den = np.hypot(x - r0[0], y - r0[1]) ** 3
    return q * (x - r0[0]) / den, q * (y - r0[1]) / den


def _angle_E(n_neg_charges):
    """Create the angle of the electric field with the x-axis"""
    # Grid of x, y points
    nx, ny = 64, 64
    x = np.linspace(-2, 2, nx)
    y = np.linspace(-2, 2, ny)
    X, Y = np.meshgrid(x, y)

    # Create a multipole with nq charges of alternating sign, equally spaced
    # on the unit circle.
    nq = 2 ** int(n_neg_charges)
    charges = []
    for i in range(nq):
        q = i % 2 * 2 - 1
        charges.append((q, (np.cos(2 * np.pi * i / nq), np.sin(2 * np.pi * i / nq))))

    # Electric field vector, E=(Ex, Ey), as separate components
    Ex, Ey = np.zeros((ny, nx)), np.zeros((ny, nx))
    for charge in charges:
        ex, ey = _E(*charge, x=X, y=Y)
        Ex += ex
        Ey += ey

    # angle wrt to 1_x
    cos_th = Ex / np.hypot(Ex, Ey)
    return np.degrees(np.arccos(cos_th)), charges, x, y, Ex, Ey


def _plot_e_field(ax, cmap, n_neg_charges=2, facecolor="black", title=False):
    """Create the mpl ax of the electric field"""
    title_color = "white" if facecolor == "black" else "black"

    th_deg, charges, x, y, Ex, Ey = _angle_E(n_neg_charges)

    # Plot the streamlines with an appropriate colormap and arrow style
    ax.streamplot(x, y, Ex, Ey, color="black", linewidth=.5,
                  density=0.5, arrowstyle='fancy', arrowsize=1)

    ang = ax.imshow(th_deg, extent=[-2, 2, -2, 2], cmap=cmap)

    # Add filled circles for the charges themselves
    charge_colors = {True: '#f50000', False: '#0091ff'}
    charge_label = {True: '+', False: '-'}
    label_align = {True: 'center', False: 'center'}
    for q, pos in charges:
        ax.add_patch(Circle(pos, 0.175, color=charge_colors[q > 0]))
        ax.annotate(charge_label[q > 0], xy=pos, fontsize=30, ha="center", va=label_align[q > 0])

    ax.set_xlabel('')
    ax.set_ylabel('')
    ax.set_xlim(-2, 2)
    ax.set_ylim(-2, 2)
    ax.set_aspect('equal')
    if title:
        ax.set_title(r"Angle of $\vec{E}$ with the x-axis", color=title_color, fontsize=14)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cax.yaxis.label.set_color(title_color)
    cax.tick_params(axis='y', colors=title_color)
    plt.colorbar(ang, cax=cax)

    return ax


def _plot_examples(color_map, images, arr_3d, figsize, facecolor, cname, cblind=True, norm=False):
    """Create the figure based on the provided images, continuous colormaps"""
    fig = plt.figure(figsize=figsize, facecolor=facecolor)
    n_images = len(images)
    if cblind:
        c_maps, sub_title = _get_color_weak_cmap(color_map, n_images)
    else:
        c_maps = [color_map]
        sub_title = [""] * n_images

    title_color = "white" if facecolor == "black" else "black"

    axi = 1
    idx_3d = 0
    n_rows = len(c_maps)
    n_cols = len(images)

    for c_map, im in itertools.product(c_maps, images):
        
        if isinstance(im, str) and ('3D' in im):
            px, py, pz = arr_3d[idx_3d]
            ax3d = fig.add_subplot(n_rows, n_cols, axi, projection="3d", facecolor=facecolor, elev=10,
                                   azim=-45)
            ax3d.plot_surface(px, py, pz, cmap=c_map, linewidth=0, antialiased=False)
            ax3d = plt.gca()
            ax3d.xaxis.set_ticklabels([])
            ax3d.yaxis.set_ticklabels([])
            ax3d.zaxis.set_ticklabels([])
            idx_3d = 0 if idx_3d == 1 else 1
        elif isinstance(im, str) and ("electric" in im):
            ax = fig.add_subplot(n_rows, n_cols, axi, facecolor=facecolor)
            ax = _plot_e_field(ax, c_map, n_neg_charges=2, facecolor=facecolor, title=axi <= n_cols)
        elif isinstance(im, str) and ("complex" in im):
            ax = fig.add_subplot(n_rows, n_cols, axi, facecolor=facecolor)
            ax = _plot_complex_arg(ax, c_map, facecolor, title=axi <= n_cols)
        else:
            cmap_norm = mpl.colors.CenteredNorm() if norm else None
            ax = fig.add_subplot(n_rows, n_cols, axi, facecolor=facecolor)
            ax.imshow(im, cmap=c_map, aspect="auto", norm=cmap_norm)
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
            ax.set_title(sub_title[axi - 1], color=title_color, fontsize=16, loc="left")

        axi += 1
    fig.suptitle(cname, color=title_color, fontsize=24, y=0.95)
    return fig


def _plot_examples_qual(color_map, dict_arr, figsize, facecolor, cname, year):
    """Create the figure with examples for discrete colormaps"""
    fig = plt.figure(figsize=figsize, facecolor=facecolor)

    c_tabs, sub_title = _get_color_weak_ctab(color_map, len(dict_arr) - 1)

    title_color = "white" if facecolor == "black" else "black"

    axi = 1
    n_rows = len(c_tabs)
    n_cols = len(dict_arr)

    for c_map, d in itertools.product(c_tabs, dict_arr):

        if axi in range(1, n_rows * n_cols, len(dict_arr)):
            ax = fig.add_subplot(n_rows, n_cols, axi, facecolor=facecolor)
            n_colors = len(d.keys())
            ax.stackplot(year, d.values(),
                         labels=d.keys(),
                         colors=c_map[range(n_colors), ...])
            ax.legend(loc='upper left')
            ax.set_facecolor(facecolor)
            ax.set_title(sub_title[axi - 1], color=title_color, fontsize=16, loc="left")
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
        elif axi in range(2, n_rows * n_cols, n_cols):
            ax = fig.add_subplot(n_rows, n_cols, axi, facecolor=facecolor)
            N = 45
            x, y = np.random.rand(2, N)
            np.random.seed(19680801)
            s = np.random.randint(10, 220, size=N)
            c = np.random.randint(0, 5, size=N)
            scatter = ax.scatter(x, y, cmap=ListedColormap(np.clip(c_map, 0, 1)), s=s, c=c)
            # produce a legend with the unique colors from the scatter
            legend1 = ax.legend(*scatter.legend_elements(),
                                loc="lower left", title="Classes")
            ax.add_artist(legend1)
            ax.set_facecolor(facecolor)
            # produce a legend with a cross section of sizes from the scatter
            handles, labels = scatter.legend_elements(prop="sizes", alpha=0.6)
            legend2 = ax.legend(handles, labels, loc="upper right", title="Sizes")
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
        else:
            ax = fig.add_subplot(n_rows, n_cols, axi, facecolor=facecolor)
            n_colors = d.shape[1]
            for col in range(n_colors):
                ax.plot(d[..., col], color=c_map[col, ...])
            ax.set_facecolor(facecolor)
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
        axi += 1
    fig.suptitle(cname, color=title_color, fontsize=24, y=0.95)
    return fig



import os.path
from setuptools import setup, find_packages

# The directory containing this file
HERE = os.path.abspath(os.path.dirname(__file__))

# The text of the README file
with open(os.path.join(HERE, "README.md")) as fid:
    README = fid.read()

EXTRAS_REQUIRE = {"tests": ["pytest", "pytest-cov"]}

INSTALL_REQUIRES = [
    "numpy",
    "pandas",
    "scipy",
    "colorspacious",
    "colorcet",
    "cmcrameri",
    "cmocean",
    "cmasher >= 1.5.8",
    "palettable >= 3.3.0",
    "matplotlib >= 3.3.0",
]

KEYWORDS = "color, color map, scientific color maps, uniform "

setup(
    name="scicomap",
    version="0.4",
    description="Scientific color maps",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Thomas Bury",
    author_email="bury.thomas@gmail.com",
    packages=find_packages(),
    zip_safe=False,  # the package can run out of an .egg file
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    python_requires=">=3.6",
    license="MIT",
    keywords=KEYWORDS,
    package_data={'': ['data/*.png', 'data/*.npz', 'data/*.gz', 'data/*.jpg']},
)



import gzip
import warnings
from pkg_resources import resource_stream
import numpy as np
import matplotlib.image as mpimg
from os.path import dirname, join


def load_hill_topography():
    """
    Load hillshading and return elevation
    :return: np.array
    """
    stream = resource_stream(__name__, 'data/jacksboro_fault_dem.npz')
    # module_path = dirname(__file__)
    # data_file_name = join(module_path, 'data', 'jacksboro_fault_dem.npz')
    with np.load(stream) as dem:
        elevation = dem["elevation"]
    return elevation


def load_scan_image():
    """
    Load image of a medical scan
    :return:
    """
    # module_path = dirname(__file__)
    # data_file_name = join(module_path, 'data', 's1045.ima.gz')
    stream = resource_stream(__name__, 'data/s1045.ima.gz')
    with gzip.open(stream) as dfile:
        scan_im = np.frombuffer(dfile.read(), np.uint16).reshape((256, 256))
    return scan_im


def load_pic(name="grmhd"):

    if not isinstance(name, str):
        TypeError("name should be a string")

    module_path = dirname(__file__)
    stream = resource_stream(__name__, 'data/grmhd.png')
    # pic_path = join(module_path, 'data', 'grmhd.png')
    if name == "grmhd":
        stream = resource_stream(__name__, 'data/grmhd.png')
    elif name == "vortex":
        stream = resource_stream(__name__, 'data/vortex.jpg')
    elif name == "tng":
        stream = resource_stream(__name__, 'data/tng.jpg')
    else:
        warnings.warn("Using a default image, name should be in ['grmhd', 'vortex', 'tng']")

    img = mpimg.imread(stream)
    return img[:, :, 0]
